// Download.cpp: implementation of the CDownload class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "Download.h"
#include "procs.h"
#include <afxinet.h>

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////
// CDownloadInfo implementation
//////////////////////////////////////////////////////////////////////

CDownloadInfo::CDownloadInfo() //default contructor
{
	parseHTMLonly = false;
}

CDownloadInfo::~CDownloadInfo() 
{
	variables.RemoveAll();
}

CDownloadInfo::CDownloadInfo(const CDownloadInfo &s ) //copy contructor
{
	filename = s.filename;
	headers = s.headers;
	method = s.method;
	object = s.object;
	path = s.path;
	port = s.port;
	server = s.server;
	parseHTMLonly = s.parseHTMLonly;
}

CDownloadInfo& CDownloadInfo::operator=(const CDownloadInfo &s ) // assignment
{
	filename = s.filename;
	headers = s.headers;
	method = s.method;
	object = s.object;
	path = s.path;
	port = s.port;
	server = s.server;
	parseHTMLonly = s.parseHTMLonly;
	return *this;
}

bool CDownloadInfo::SetURL(CString url)
{
	DWORD service;
	return (AfxParseURL(url, service, server, object, port)!=0);
}

void CDownloadInfo::SetOptions(int method, CString headers, CString filename, CString path)
{
	SetMethod(method);
	this->headers = headers;
	this->filename = filename;
	this->path = path;
}

void CDownloadInfo::SetMethod(int method) {
	this->method = method;
}

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

IMPLEMENT_DYNCREATE(CDownload, CObject)

CDownload::CDownload(CDownloadCallback* callback)
{
	this->session = NULL;
	this->connection = NULL;
	this->callback = callback;
}

CDownload::~CDownload()
{
	CloseSession();
}

BOOL CDownload::OpenSession(CString proxy, int retries, int timeout_c, int timeout_r, int timeout_s) {
	try {
		DWORD dwAccessType=INTERNET_OPEN_TYPE_DIRECT;
		this->retries=retries;
		if (proxy!="") {
			dwAccessType=INTERNET_OPEN_TYPE_PROXY;
		}
		session = new CInternetSession(
			"Map Grabber", 1,
			dwAccessType,
			proxy,
			NULL,
			INTERNET_FLAG_DONT_CACHE);
		if (timeout_s != 0) {
			if (!session->SetOption(INTERNET_OPTION_CONNECT_TIMEOUT, timeout_c) ||
				!session->SetOption(INTERNET_OPTION_CONNECT_RETRIES, retries) ||
				!session->SetOption(INTERNET_OPTION_DATA_SEND_TIMEOUT, timeout_s) ||
				!session->SetOption(INTERNET_OPTION_DATA_RECEIVE_TIMEOUT, timeout_r)) {
				errormsg = "Error opening session [SetOption()]";
				CloseSession();
				return 1;
			}
		}
	}
	catch (CInternetException* pEx){
		pEx->GetErrorMessage(errormsg.GetBufferSetLength(255), 255);
		errormsg.ReleaseBuffer();
		pEx->Delete();
		errormsg = "Error opening session\n - " + errormsg;
		CloseSession();
		return 1;
	}
	return 0;
}

BOOL CDownload::OpenConnection(CDownloadInfo* info) {
	if (session == NULL) return 1;
	try {
		connection = session->GetHttpConnection(info->server,
			info->port);
	}
	catch (CInternetException* pEx){
		pEx->GetErrorMessage(errormsg.GetBufferSetLength(255), 255);
		errormsg.ReleaseBuffer();
		errormsg = "Error opening connection :\n" + errormsg;
		pEx->Delete();
		CloseSession();
		return 1;
	}
	return 0;
}

// DOWNLOAD A FILE
int CDownload::Start(CDownloadInfo* info)
{
	const bufsize = 8192;
	char buffer[bufsize];
	unsigned long errcode;
	int actread, size, total, verb;
	CString objectname, sizestr, query, optional;
	POSITION pos;
	int own_retries;
	BOOL can_retry;

	objectname = info->object;

// VARIABLES FOR DIFFERENT METHODS
	switch (info->method) {
		case METHOD_NONE:
			verb = CHttpConnection::HTTP_VERB_GET;
			break;
		case METHOD_GET:
		case METHOD_POST:
			pos = info->variables.GetHeadPosition();
			while (pos != NULL) {
				Tvariable var = info->variables.GetNext(pos);
				query += var.name+"="+var.value;
				if (pos != NULL) query += "&";
			}
			switch (info->method) {
				case METHOD_GET:
					objectname += "?"+query;
					verb = CHttpConnection::HTTP_VERB_GET;
					break;
				case METHOD_POST:
					optional = URLEncode(query);
					verb = CHttpConnection::HTTP_VERB_POST;
					break;
			}
			break;
	}
// CLOSE CONNECTION IF THE SERVER HAS CHANGED
	if (last_server != info->server) CloseConnection();
	last_server = info->server;
// OPEN CONNECTION IF NOT ALREADY OPEN
	if (connection == NULL) {
		if (OpenConnection(info)!=0) return 1;
	}
// MANUAL RETRY LOOP
	own_retries=retries;
	can_retry=true;
	while (can_retry) { // manual "continue"
		own_retries--;
		can_retry=false;

// CREATING REQUEST
		CHttpFile* request = connection->
			OpenRequest(verb, objectname, NULL, 1, NULL, "HTTP/1.0",
			INTERNET_FLAG_EXISTING_CONNECT | INTERNET_FLAG_DONT_CACHE);
		try {
			size = 0; total = 1;
			if (callback) {
				if (!callback->DownloadProgress(size, total)) {
					// order to be killed
					request->Close();
					delete request;
					return 2;
				}
			}
			request->SendRequest(info->headers,
				optional.GetBuffer(0), optional.GetLength());
			request->QueryInfoStatusCode(errcode);
			request->QueryInfo(HTTP_QUERY_CONTENT_LENGTH, sizestr);
			total = strtol(sizestr, NULL, 10 );
			if (errcode == HTTP_STATUS_OK) {
				CFile file;
				CFileException e;
				CHtml html;
				CString filename = info->path+info->filename;
				CString tmpfilename = info->path+"$"+info->filename;
				if (info->parseHTMLonly ||
					file.Open(tmpfilename, CFile::modeCreate | CFile::modeWrite, &e)) {
					actread = bufsize;
					while (actread > 0) {
						actread = request->Read(buffer, bufsize);
						if (actread > 0) {
							if (info->parseHTMLonly)
								html.ParseBuffer(buffer, actread, this);
							else
								file.Write(buffer, actread);
						}
						size += actread;
						if (callback) {
							if (!callback->DownloadProgress(size, total)) {
								// order to be killed
								if (!info->parseHTMLonly) file.Close();
								request->Close();
								delete request;
								return 2;
							}
						}

					}
					if (!info->parseHTMLonly) {
						file.Close(); // successful download
						try {
							if (FileExist(filename)) CFile::Remove(filename);
							CFile::Rename(tmpfilename, filename);
						}
						catch (CFileException* pEx) {
							pEx->GetErrorMessage(errormsg.GetBufferSetLength(255), 255);
							errormsg.ReleaseBuffer();
							errormsg = "Error renaming "+tmpfilename+
								" to "+filename+" :\n"+errormsg;
							pEx->Delete();
							request->Close();
							delete request;
							return 1;
						}
					}
				}
				else {
					e.GetErrorMessage(errormsg.GetBufferSetLength(255), 255);
					errormsg.ReleaseBuffer();
					errormsg = "Error downloading file:\n"+errormsg;
					request->Close();
					delete request;
					return 1;
				}
			}
			else {
				errormsg = "Error downloading "+objectname+"\n\nHTTP status code : "+Str(errcode)+" ";
				switch (errcode) {
					case 204: errormsg += "(No content)"; break;
					case 404: errormsg += "(Object not found)"; break;
					case 301: case 302: errormsg += "(Object moved)"; break;
					case 401: errormsg += "(Access denied)"; break;
					case 403: errormsg += "(Forbidden)"; break;
					case 408: errormsg += "(Request timeout)"; break;
					case 500: errormsg += "(Internal server error)"; break;
					case 503: errormsg += "(Temporarily overloaded)"; break;
				}
				request->Close();
				delete request;
				return 1;
			}
			request->Close();
			delete request;
		}
		catch (CInternetException* pEx){
			pEx->GetErrorMessage(errormsg.GetBufferSetLength(255), 255);
			errormsg.ReleaseBuffer();
			errormsg = "Error downloading "+objectname+"\n\n"+errormsg;
			int errorcode = pEx->m_dwError & 0xFFFF;
			pEx->Delete();
			request->Close();
			delete request;
			if (own_retries>0) {
				can_retry = true;
				continue;
			}
			return 1;
		}
	}
	return 0;
}

void CDownload::ProcessHTML(CString tag, CString par, CString value)
{
	if (callback) callback->ProcessHTML(tag, par, value);
}

void CDownload::CloseConnection()
{
	if (connection!=NULL) {
		connection->Close();
		delete connection;
		connection = NULL;
	}
}

void CDownload::CloseSession()
{
	CloseConnection();
	if (session!=NULL) {
		session->Close();
		delete session;
		session = NULL;
	}
}

CString CDownload::GetError()
{
	return errormsg;
}
